import asyncio
from dataclasses import dataclass
from enum import Enum
import logging
from typing import (
    Any,
    cast,
    Coroutine,
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

from uniswap_universal_router_decoder import RouterCodec
from web3 import AsyncWeb3
from web3.contract import AsyncContract
from web3.exceptions import Web3Exception
from web3.types import (
    ChecksumAddress,
    Wei,
)

from ._utilities import to_wei
from .smart_rate_limiter import (
    _rate_limit,
    SmartRateLimiter,
)


logger = logging.getLogger(__name__)
codec = RouterCodec()


@dataclass(frozen=True)
class Token:
    address: ChecksumAddress
    symbol: str
    decimals: int


class RouterFunction(Enum):
    V3_SWAP_EXACT_IN = "V3_SWAP_EXACT_IN"
    V3_SWAP_EXACT_OUT = "V3_SWAP_EXACT_OUT"
    V2_SWAP_EXACT_IN = "V2_SWAP_EXACT_IN"
    V2_SWAP_EXACT_OUT = "V2_SWAP_EXACT_OUT"


@dataclass(frozen=True)
class V2OrderedPool:
    token_in: Token
    token_out: Token


@dataclass(frozen=True)
class V3OrderedPool:
    token_in: Token
    pool_fee: int
    token_out: Token


OrderedPool = TypeVar('OrderedPool', V2OrderedPool, V3OrderedPool)
V2PathList = Tuple[ChecksumAddress, ...]
V3PathList = Tuple[Union[int, ChecksumAddress], ...]
PathList = TypeVar('PathList', V2PathList, V3PathList)


class PoolPath(Protocol[OrderedPool, PathList]):
    pools: Sequence[OrderedPool]
    def get_path(self) -> PathList: ...
    def to_dict(self) -> Dict[str, PathList]: ...
    async def get_amount_out(self, amount_in: Wei) -> Wei: ...
    def get_smart_rate_limiter(self) -> Optional[SmartRateLimiter]: ...


class V2PoolPath(PoolPath[V2OrderedPool, V2PathList]):
    contract: AsyncContract = AsyncWeb3().eth.contract(AsyncWeb3.to_checksum_address("0" * 40))

    def __init__(self, pools: Sequence[V2OrderedPool], smart_rate_limiter: Optional[SmartRateLimiter] = None) -> None:
        self.pools = pools
        self.path = self._build_path()
        self.smart_rate_limiter = smart_rate_limiter

    def get_smart_rate_limiter(self) -> Optional[SmartRateLimiter]:
        return self.smart_rate_limiter

    def get_path(self) -> V2PathList:
        return self.path

    def _build_path(self) -> V2PathList:
        path = [self.pools[0].token_in.address]
        for pool in self.pools:
            path.append(pool.token_out.address)
        return tuple(path)

    def to_dict(self) -> Dict[str, V2PathList]:
        return {"path": self.get_path()}

    @_rate_limit("eth_call")
    async def get_amount_out(self, amount_in: Wei) -> Wei:
        quote = await self.contract.functions.getAmountsOut(amount_in, self.get_path()).call()
        return to_wei(quote[-1])

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.path}"


class V3PoolPath(PoolPath[V3OrderedPool, V3PathList]):
    contract: AsyncContract = AsyncWeb3().eth.contract(AsyncWeb3.to_checksum_address("0" * 40))

    def __init__(self, pools: Sequence[V3OrderedPool], smart_rate_limiter: Optional[SmartRateLimiter] = None) -> None:
        self.pools = pools
        self.path = self._build_path()
        self.smart_rate_limiter = smart_rate_limiter

    def get_smart_rate_limiter(self) -> Optional[SmartRateLimiter]:
        return self.smart_rate_limiter

    def get_path(self) -> V3PathList:
        return self.path

    def _build_path(self) -> V3PathList:
        path: List[Union[int, ChecksumAddress]] = [self.pools[0].token_in.address]
        for pool in self.pools:
            path.append(pool.pool_fee)
            path.append(pool.token_out.address)
        return tuple(path)

    def to_dict(self) -> Dict[str, V3PathList]:
        return {"path": self.get_path()}

    @_rate_limit("eth_call")
    async def get_amount_out(self, amount_in: Wei) -> Wei:
        encoded_path = codec.encode.v3_path("V3_SWAP_EXACT_IN", self.get_path())
        quote = await self.contract.functions.quoteExactInput(encoded_path, amount_in).call()
        return to_wei(quote[0])

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.path}"


@dataclass(frozen=True)
class WeightedPath:
    router_function: RouterFunction
    pool_path: PoolPath  # type: ignore
    weight: int

    def to_dict(self) -> Dict[str, Union[str, PathList, int]]:
        result = self.pool_path.to_dict()
        result["function"] = self.router_function.value
        result["weight"] = self.weight
        return result


class WeightedPathResult(TypedDict):
    function: str
    path: PathList  # type: ignore
    weight: int
    estimate: Wei


class MixedWeightedPath:
    def __init__(self, weighted_paths: Sequence[WeightedPath]) -> None:
        self.weighted_paths: Tuple[WeightedPath, ...] = tuple(weighted_paths)
        self.values: Tuple[Wei, ...] = (Wei(0), Wei(0))
        self.total_value: Wei = Wei(0)

    async def compute_path_values(self, amount: Wei) -> None:
        computing_coros: List[Coroutine[Any, Any, Wei]] = [
            w_p.pool_path.get_amount_out(Wei(amount * w_p.weight // 100))
            for w_p in self.weighted_paths
        ]
        try:
            self.values = tuple(await asyncio.gather(*computing_coros))
            self.total_value = Wei(sum(self.values))
        except (asyncio.exceptions.TimeoutError, ValueError, Web3Exception) as e:
            logger.debug(f"Could not compute value for path(s): {self.weighted_paths}. Reason: {e}")

    def output(self) -> Tuple[WeightedPathResult, ...]:
        output = []
        for i, path in enumerate(self.weighted_paths):
            path_dict = path.to_dict()
            path_dict["estimate"] = self.values[i]
            output.append(cast(WeightedPathResult, path_dict))

        return tuple(output)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: {self.weighted_paths}, "
            f"values: {self.values}, "
            f"total_value: {self.total_value}"
        )



import pytest
from web3.types import Wei

from uniswap_smart_path import SmartPath
from uniswap_smart_path._datastructures import (  # noqa
    MixedWeightedPath,
    RouterFunction,
    Token,
    V2OrderedPool,
    V2PoolPath,
    V3OrderedPool,
    V3PoolPath,
    WeightedPath,
)

from .conftest import tokens


@pytest.mark.parametrize(
    "pools, amount_in, expected_path, expected_amount",
    (
            ((V2OrderedPool(tokens["USDC"], tokens["DAI"]), ), 1000 * 10**6, (tokens["USDC"].address, tokens["DAI"].address), 1000 * 10**18),  # noqa
            ((V2OrderedPool(tokens["USDC"], tokens["DAI"]), V2OrderedPool(tokens["DAI"], tokens["USDT"])), 1000 * 10 ** 6, (tokens["USDC"].address, tokens["DAI"].address, tokens["USDT"].address), 1000 * 10 ** 6),  # noqa
    )
)
async def test_v2_pool_path(pools, amount_in, expected_path, expected_amount, w3):
    await SmartPath.create(w3)
    v2_pool_path = V2PoolPath(pools)

    assert expected_path == v2_pool_path.get_path() == v2_pool_path.to_dict()["path"]
    assert 0.93 * expected_amount < await v2_pool_path.get_amount_out(amount_in) < 1.07 * expected_amount


@pytest.mark.parametrize(
    "pools, amount_in, expected_path, expected_amount",
    (
            ((V3OrderedPool(tokens["USDC"], 500, tokens["DAI"]), ), 1000 * 10**6, (tokens["USDC"].address, 500, tokens["DAI"].address), 1000 * 10**18),  # noqa
            ((V3OrderedPool(tokens["USDC"], 100, tokens["DAI"]), V3OrderedPool(tokens["DAI"], 100, tokens["USDT"])), 1000 * 10 ** 6, (tokens["USDC"].address, 100, tokens["DAI"].address, 100, tokens["USDT"].address), 1000 * 10 ** 6),  # noqa
    )
)
async def test_v3_pool_path(pools, amount_in, expected_path, expected_amount, w3):
    await SmartPath.create(w3)
    v3_pool_path = V3PoolPath(pools)

    assert expected_path == v3_pool_path.get_path() == v3_pool_path.to_dict()["path"]
    assert 0.95 * expected_amount < await v3_pool_path.get_amount_out(amount_in) < 1.05 * expected_amount


v2_pool_path_1 = V2PoolPath([V2OrderedPool(tokens["DAI"], tokens["USDC"])])
weighted_path_1 = WeightedPath(RouterFunction.V2_SWAP_EXACT_IN, v2_pool_path_1, 40)

v3_pool_path_2 = V3PoolPath([V3OrderedPool(tokens["DAI"], 500, tokens["USDC"])])
weighted_path_2 = WeightedPath(RouterFunction.V3_SWAP_EXACT_IN, v3_pool_path_2, 60)


async def test_mixed_weighted_path(w3):
    await SmartPath.create(w3)
    mixed_path = MixedWeightedPath([weighted_path_1, weighted_path_2])
    await mixed_path.compute_path_values(Wei(100 * 10**18))
    assert 40 * 10**6 * 0.97 < mixed_path.values[0] < 40 * 10**6 * 1.03
    assert 60 * 10 ** 6 * 0.97 < mixed_path.values[1] < 60 * 10 ** 6 * 1.03
    assert 100 * 10 ** 6 * 0.97 < mixed_path.total_value < 100 * 10 ** 6 * 1.03
    assert "V2_SWAP_EXACT_IN" in str(mixed_path)
    assert "V3_SWAP_EXACT_IN" in str(mixed_path)

    # test compute_path_values except statement
    await mixed_path.compute_path_values(-1)  # noqa



import pytest
from web3 import Web3
from web3.exceptions import BadFunctionCallOutput
from web3.types import Wei

from uniswap_smart_path import (
    SmartPath,
    SmartRateLimiter,
)
import uniswap_smart_path._constants as const  # noqa
from uniswap_smart_path._datastructures import (  # noqa
    MixedWeightedPath,
    RouterFunction,
    Token,
    V2OrderedPool,
    V2PoolPath,
    V3OrderedPool,
    V3PoolPath,
    WeightedPath,
)
from uniswap_smart_path.exceptions import SmartPathException

from .conftest import tokens


credit_limiter = SmartRateLimiter(1, max_credits=40, method_credits={"eth_call": 20})
count_limiter = SmartRateLimiter(1, max_count=2)


async def test_create(w3, rpc_endpoint, uniswapv2_address, uniswapv3_quoter_address):
    sp_init_w3 = SmartPath(w3=w3)

    assert V2PoolPath.contract.address == uniswapv2_address == sp_init_w3.uniswapv2.address
    assert V3PoolPath.contract.address == uniswapv3_quoter_address == sp_init_w3.quoter.address

    sp_create_w3 = await SmartPath.create(w3=w3)
    sp_create_rpc = await SmartPath.create(rpc_endpoint=rpc_endpoint)

    assert sp_init_w3.chain_id == sp_create_w3.chain_id == sp_create_rpc.chain_id == 1

    assert len(sp_create_w3.v3_pools_fees_x_pivots) == 16
    assert (tokens["WETH"], 3000) in sp_create_w3.v3_pools_fees_x_pivots

    with pytest.raises(ValueError):
        _ = await SmartPath.create()

    with pytest.raises(ValueError):
        _ = SmartRateLimiter(1, max_credits=40)

    with pytest.raises(ValueError):
        _ = SmartRateLimiter(1)

    with pytest.raises(NotImplementedError):
        _ = await SmartPath.create(w3=w3, with_gas_estimate=True)


@pytest.mark.parametrize(
    "token, smart_rate_limiter, expected_exception",
    (
        (tokens["WETH"], None, None),
        (tokens["WETH"], credit_limiter, None),
        # (tokens["WETH"], count_limiter, None),  # issue between pytest-asyncio and python 3.8 & 3.9  # noqa
        (tokens["MKR"], None, None),
        (tokens["MKR"], credit_limiter, None),
        # (tokens["MKR"], count_limiter, None),  # issue between pytest-asyncio and python 3.8 & 3.9  # noqa
        (tokens["FAKE"], None, BadFunctionCallOutput),  # BadFunctionCallOutput because of decimals call
        (Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), None, BadFunctionCallOutput)
    )
)
async def test_get_token(token, smart_rate_limiter, expected_exception, w3):
    smart_path = await SmartPath.create(w3, smart_rate_limiter=smart_rate_limiter)
    if expected_exception:
        with pytest.raises(expected_exception):
            _ = await smart_path._get_token(token.address, w3)
    else:
        assert token == await smart_path._get_token(token.address, w3)


@pytest.mark.parametrize(
    "token0, token1, smart_rate_limiter, expected_result",
    (
        (tokens["WETH"], tokens["USDC"], None, True),
        (tokens["WETH"], tokens["USDC"], credit_limiter, True),
        # (tokens["WETH"], tokens["USDC"], count_limiter, True),  # issue between pytest-asyncio and python 3.8 & 3.9  # noqa
        (tokens["WETH"], tokens["FAKE"], None, False),
        (tokens["WETH"], tokens["FAKE"], credit_limiter, False),
        # (tokens["WETH"], tokens["FAKE"], count_limiter, False),  # issue between pytest-asyncio and python 3.8 & 3.9  # noqa
        (tokens["WETH"], Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), None, False),
        (tokens["WETH"], Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), credit_limiter, False),
        (tokens["WETH"], Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), count_limiter, False),
        (Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), tokens["USDC"], None, False),
    )
)
async def test_v2_pool_exist(token0, token1, smart_rate_limiter, expected_result, w3):
    smart_path = await SmartPath.create(w3, smart_rate_limiter=smart_rate_limiter)
    assert expected_result == await smart_path._v2_pool_exist(token0, token1)


@pytest.mark.parametrize(
    "token0, token1, fees, smart_rate_limiter, expected_result",
    (
        (tokens["WETH"], tokens["USDC"], 3000, None, True),
        (tokens["WETH"], tokens["USDC"], 3000, credit_limiter, True),
        (tokens["WETH"], tokens["USDC"], 3000, count_limiter, True),
        (tokens["WETH"], tokens["FAKE"], 3000, None, False),
        (tokens["WETH"], Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), 3000, None, False),
        (Token(Web3.to_checksum_address("0" * 40), "NotAToken", 0), tokens["USDC"], 3000, None, False),
    )
)
async def test_v3_pool_exist(token0, token1, fees, smart_rate_limiter, expected_result, w3):
    smart_path = await SmartPath.create(w3, smart_rate_limiter=smart_rate_limiter)
    assert expected_result == await smart_path._v3_pool_exist(token0, token1, fees)


@pytest.mark.parametrize(
    "token0, token1, pivot, smart_rate_limiter, expected_result",
    (
        (tokens["DAI"], tokens["WETH"], tokens["USDC"], None, True),
        (tokens["DAI"], tokens["WETH"], tokens["USDC"], credit_limiter, True),
        (tokens["DAI"], tokens["WETH"], tokens["FAKE"], None, False),
        (tokens["DAI"], tokens["WETH"], tokens["FAKE"], count_limiter, False),
    )
)
async def test_v2_pools_exists_for_pivot_token(token0, token1, pivot, smart_rate_limiter, expected_result, w3):
    smart_path = await SmartPath.create(w3, smart_rate_limiter=smart_rate_limiter)
    assert expected_result == await smart_path._v2_pools_exists_for_pivot_token(token0, token1, pivot)


expected_v2_path_list_01 = [
    ('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48'),
    ('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', '0xdAC17F958D2ee523a2206206994597C13D831ec7', '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48'),  # noqa
    ('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', '0x6B175474E89094C44Da98b954EedeAC495271d0F', '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48'),  # noqa
]
expected_v2_path_list_02 = [
    ('0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984', '0x514910771AF9Ca656af840dff83E8264EcF986CA'),
    ('0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984', '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48', '0x514910771AF9Ca656af840dff83E8264EcF986CA'),  # noqa
    ('0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984', '0xdAC17F958D2ee523a2206206994597C13D831ec7', '0x514910771AF9Ca656af840dff83E8264EcF986CA'),  # noqa
    ('0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984', '0x6B175474E89094C44Da98b954EedeAC495271d0F', '0x514910771AF9Ca656af840dff83E8264EcF986CA'),  # noqa
    ('0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984', '0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', '0x514910771AF9Ca656af840dff83E8264EcF986CA'),  # noqa
]


@pytest.mark.parametrize(
    "token_in, token_out, expected_result",
    (
        (tokens["WETH"], tokens["USDC"], expected_v2_path_list_01),
        (tokens["UNI"], tokens["LINK"], expected_v2_path_list_02),
        (tokens["FAKE"], tokens["WETH"], []),
    )
)
async def test_build_v2_path_list(token_in, token_out, expected_result, w3):
    smart_path = await SmartPath.create(w3)
    v2_pool_path_list = await smart_path._build_v2_path_list(token_in, token_out)
    assert len(v2_pool_path_list) == len(expected_result)
    for pool in v2_pool_path_list:
        assert pool.get_path() in expected_result


@pytest.mark.parametrize(
    "token_in, is_token_in, expected_number_of_results",
    (
        (tokens["USDC"], True, 12),
        (tokens["WETH"], True, 12),
        (tokens["UNI"], False, 13),
        (tokens["FAKE"], True, 0),
    )
)
async def test_build_v3_base_pool_list(token_in, is_token_in, expected_number_of_results, w3):
    smart_path = await SmartPath.create(w3)
    v3_base_pool_list = await smart_path._get_v3_base_pools(token_in, is_token_in)
    assert len(v3_base_pool_list) == expected_number_of_results
    for pool in v3_base_pool_list:
        assert pool.token_in != pool.token_out
        assert token_in == pool.token_in if is_token_in else pool.token_out


@pytest.mark.parametrize(
    "token_in, token_out, expected_number_of_results, expected_fees",
    (
        (tokens["WETH"], tokens["USDC"], 4, (100, 500, 3000, 10000)),
        (tokens["UNI"], tokens["LINK"], 2, (3000, 10000)),
        (tokens["WETH"], tokens["FAKE"], 0, ()),
    )
)
async def test_get_v3_one_hop_pools(token_in, token_out, expected_number_of_results, expected_fees, w3):
    smart_path = await SmartPath.create(w3)
    v3_one_hop_pools = await smart_path._get_v3_one_hop_pools(token_in, token_out)
    assert len(v3_one_hop_pools) == expected_number_of_results
    for pool in v3_one_hop_pools:
        assert token_in == pool.token_in
        assert token_out == pool.token_out
        assert pool.pool_fee in expected_fees


@pytest.mark.parametrize(
    "token_in, token_out, expected_number_of_results",
    (
        (tokens["WETH"], tokens["USDC"], 36),
        (tokens["UNI"], tokens["LINK"], 35),
    )
)
async def test_build_v3_path_list(token_in, token_out, expected_number_of_results, w3):
    smart_path = await SmartPath.create(w3)
    v3_pool_paths = await smart_path._build_v3_path_list(token_in, token_out)
    assert len(v3_pool_paths) == expected_number_of_results
    for pool in v3_pool_paths:
        assert pool.pools[0].token_in != pool.pools[0].token_out
        assert pool.pools[-1].token_in != pool.pools[-1].token_out
        pool_path = pool.get_path()
        assert pool_path[0] == token_in.address
        assert pool_path[-1] == token_out.address


v2_pool_path_1 = V2PoolPath([V2OrderedPool(tokens["WETH"], tokens["USDC"])])
weighted_path_1 = WeightedPath(RouterFunction.V2_SWAP_EXACT_IN, v2_pool_path_1, 100)
mixed_path_1 = MixedWeightedPath([weighted_path_1, ])
best_value = 1800 * 10**6
mixed_path_1.total_value = Wei(best_value)

v3_pool_path_2 = V3PoolPath([V3OrderedPool(tokens["WETH"], 3000, tokens["USDC"])])
weighted_path_2 = WeightedPath(RouterFunction.V3_SWAP_EXACT_IN, v3_pool_path_2, 100)
mixed_path_2 = MixedWeightedPath([weighted_path_2, ])
mixed_path_2.total_value = Wei(int(best_value * (const.irrelevant_value_filter_multiplier - 0.1)))

v3_pool_path_3 = V3PoolPath([V3OrderedPool(tokens["WETH"], 500, tokens["USDC"])])
weighted_path_3 = WeightedPath(RouterFunction.V3_SWAP_EXACT_IN, v3_pool_path_3, 100)
mixed_path_3 = MixedWeightedPath([weighted_path_3, ])
mixed_path_3.total_value = Wei(int(best_value * (const.irrelevant_value_filter_multiplier + 0.1)))

mixed_weighted_paths = [mixed_path_1, mixed_path_2, mixed_path_3]
expected_mixed_weighted_paths = [mixed_path_1, mixed_path_3]


def test_filter_irrelevant_low_values():
    assert expected_mixed_weighted_paths == SmartPath._filter_irrelevant_low_values(mixed_weighted_paths, best_value)


def test_get_all_mixed_path():
    all_mixed_paths = SmartPath._get_all_mixed_path(mixed_path_3, mixed_path_1)
    for i, mixed_path in enumerate(all_mixed_paths):
        assert mixed_path.weighted_paths[0].router_function == RouterFunction.V3_SWAP_EXACT_IN
        assert mixed_path.weighted_paths[0].pool_path == v3_pool_path_3
        assert mixed_path.weighted_paths[0].weight == const.weight_combinations[i][0]

        assert mixed_path.weighted_paths[1].router_function == RouterFunction.V2_SWAP_EXACT_IN
        assert mixed_path.weighted_paths[1].pool_path == v2_pool_path_1
        assert mixed_path.weighted_paths[1].weight == const.weight_combinations[i][1]


async def perform_get_swap_in_path_tests(amount, expected_estimate, smart_path, token_in, token_out):
    weighted_paths = await smart_path.get_swap_in_path(amount, token_in.address, token_out.address)
    total_estimate = 0
    for path in weighted_paths:
        total_estimate += path["estimate"]
    if expected_estimate:
        assert expected_estimate * 0.98 < total_estimate < expected_estimate * 1.02
    else:
        assert weighted_paths == ()


@pytest.mark.parametrize(
    "amount, token_in, token_out, expected_estimate",
    (
        (Wei(100 * 10**18), tokens["DAI"], tokens["USDT"], 100 * 10**6),
        (Wei(100 * 10**18), Token(Web3.to_checksum_address("0x1fB90FFC02D01238Cd8AFE3a82B8C65BAC37042f"), "", 18), tokens["USDT"], None),  # noqa
    )
)
async def test_get_swap_in_path(amount, token_in, token_out, expected_estimate, w3):
    smart_path = await SmartPath.create(w3)
    await perform_get_swap_in_path_tests(amount, expected_estimate, smart_path, token_in, token_out)

    pivots = [token.address for token in const.pivot_tokens[1]]
    custom_smart_path = await SmartPath.create_custom(
        w3,
        pivot_tokens=pivots,
        v3_pool_fees=(100, 500, 3000, 10000),
        v2_router=const.uniswapv2_address,
        v2_factory=const.uniswapv2_factory_address,
        v3_quoter=const.uniswapv3_quoter_address,
        v3_factory=const.uniswapv3_factory_address,
    )
    await perform_get_swap_in_path_tests(amount, expected_estimate, custom_smart_path, token_in, token_out)


@pytest.mark.parametrize(
    "amount, token_in, token_out, smart_rate_limiter, expected_estimate",
    (
            (Wei(100 * 10 ** 18), tokens["DAI"], tokens["USDT"], None, 100 * 10 ** 6),
            (Wei(100 * 10 ** 18), tokens["DAI"], tokens["USDT"], credit_limiter, 100 * 10 ** 6),
            # (Wei(100 * 10 ** 18), tokens["DAI"], tokens["USDT"], count_limiter, 100 * 10 ** 6),  # issue between pytest-asyncio and python 3.8 & 3.9  # noqa
            (Wei(100 * 10 ** 18), Token(Web3.to_checksum_address("0x1fB90FFC02D01238Cd8AFE3a82B8C65BAC37042f"), "", 18), tokens["USDT"], None, None),  # noqa
    )
)
async def test_get_swap_in_path_v2_only(amount, token_in, token_out, smart_rate_limiter, expected_estimate, w3):
    smart_path = await SmartPath.create_v2_only(w3, smart_rate_limiter=smart_rate_limiter)
    assert smart_rate_limiter == smart_path.smart_rate_limiter
    await perform_get_swap_in_path_tests(amount, expected_estimate, smart_path, token_in, token_out)

    custom_smart_path = await SmartPath.create_custom(
        w3,
        v2_router=const.uniswapv2_address,
        v2_factory=const.uniswapv2_factory_address,
        smart_rate_limiter=smart_rate_limiter,
    )
    assert smart_rate_limiter == custom_smart_path.smart_rate_limiter
    await perform_get_swap_in_path_tests(amount, expected_estimate, custom_smart_path, token_in, token_out)


@pytest.mark.parametrize(
    "amount, token_in, token_out, smart_rate_limiter, expected_estimate",
    (
            (Wei(100 * 10 ** 18), tokens["DAI"], tokens["USDT"], None, 100 * 10 ** 6),
            (Wei(100 * 10 ** 18), tokens["DAI"], tokens["USDT"], credit_limiter, 100 * 10 ** 6),
            # (Wei(100 * 10 ** 18), tokens["DAI"], tokens["USDT"], count_limiter, 100 * 10 ** 6),  # issue between pytest-asyncio and python 3.8 & 3.9  # noqa
            (Wei(100 * 10 ** 18), Token(Web3.to_checksum_address("0x1fB90FFC02D01238Cd8AFE3a82B8C65BAC37042f"), "", 18), tokens["USDT"], None, None),  # noqa
    )
)
async def test_get_swap_in_path_v3_only(amount, token_in, token_out, smart_rate_limiter, expected_estimate, w3):
    smart_path = await SmartPath.create_v3_only(w3, smart_rate_limiter=smart_rate_limiter)
    assert smart_rate_limiter == smart_path.smart_rate_limiter
    await perform_get_swap_in_path_tests(amount, expected_estimate, smart_path, token_in, token_out)

    custom_smart_path = await SmartPath.create_custom(
        w3,
        v3_quoter=const.uniswapv3_quoter_address,
        v3_factory=const.uniswapv3_factory_address,
        smart_rate_limiter=smart_rate_limiter,
    )
    assert smart_rate_limiter == custom_smart_path.smart_rate_limiter
    await perform_get_swap_in_path_tests(amount, expected_estimate, custom_smart_path, token_in, token_out)


@pytest.mark.parametrize(
    "pivots, expected_pivot_tokens",
    (
        ((tokens["USDC"].address, tokens["WETH"].address), (tokens["USDC"], tokens["WETH"])),
        ((), ()),
    )
)
async def test_get_pivot_tokens(pivots, expected_pivot_tokens, w3):
    smart_path = await SmartPath.create(w3)
    assert await smart_path._get_pivot_tokens(pivots, w3) == expected_pivot_tokens


async def test_create_custom_exception(w3):
    with pytest.raises(SmartPathException):
        _ = await SmartPath.create_custom(
            w3,
            v3_factory=const.uniswapv3_factory_address,
        )

    with pytest.raises(SmartPathException):
        _ = await SmartPath.create_custom(
            w3,
            v3_quoter=const.uniswapv3_quoter_address,
        )

    with pytest.raises(SmartPathException):
        _ = await SmartPath.create_custom(
            w3,
            v2_factory=const.uniswapv2_factory_address,
        )

    with pytest.raises(SmartPathException):
        _ = await SmartPath.create_custom(
            w3,
            v2_router=const.uniswapv2_address,
        )



import asyncio
import itertools
import logging
from typing import (
    Any,
    Awaitable,
    cast,
    List,
    Optional,
    Sequence,
    Tuple,
)

from web3 import (
    AsyncHTTPProvider,
    AsyncWeb3,
)
from web3.contract import AsyncContract
from web3.contract.async_contract import AsyncContractFunctions
from web3.exceptions import BadFunctionCallOutput
from web3.middleware import validation
from web3.types import (
    ChecksumAddress,
    RPCEndpoint,
    Wei,
)

from ._constants import (
    erc20_abi,
    irrelevant_value_filter_multiplier,
    pivot_tokens,
    uniswapv2_abi,
    uniswapv2_address,
    uniswapv2_factory_abi,
    uniswapv2_factory_address,
    uniswapv3_factory_abi,
    uniswapv3_factory_address,
    uniswapv3_quoter_abi,
    uniswapv3_quoter_address,
    v3_pool_fees,
    weight_combinations,
)
from ._datastructures import (
    MixedWeightedPath,
    RouterFunction,
    Token,
    V2OrderedPool,
    V2PoolPath,
    V3OrderedPool,
    V3PoolPath,
    WeightedPath,
    WeightedPathResult,
)
from ._utilities import is_null_address
from .exceptions import SmartPathException
from .smart_rate_limiter import (
    _rate_limit,
    SmartRateLimiter,
)


logger = logging.getLogger(__name__)


NO_VALIDATION_METHODS = [RPCEndpoint("eth_call")]  # to avoid unnecessary eth_chainId requests


class SmartPath:
    def __init__(
            self,
            w3: AsyncWeb3,
            with_gas_estimate: bool = False,
            chain_id: int = 1,
            with_v2: bool = True,
            with_v3: bool = True,
            smart_rate_limiter: Optional[SmartRateLimiter] = None,
            **kwargs: Any) -> None:
        if with_gas_estimate:
            raise NotImplementedError("Gas is not yet estimated")
        self.w3 = w3
        for method in NO_VALIDATION_METHODS:
            if method in validation.METHODS_TO_VALIDATE:
                logger.debug(f"Removing {method} from web3.middleware.validation.METHODS_TO_VALIDATE")
                validation.METHODS_TO_VALIDATE.remove(method)
        self.chain_id = chain_id
        self.with_v2 = with_v2
        self.with_v3 = with_v3

        self.pivots = kwargs.get("pivot_tokens") or pivot_tokens[self.chain_id]

        if self.with_v2:
            v2_router = w3.to_checksum_address(kwargs.get("v2_router") or uniswapv2_address)
            self.uniswapv2 = self.w3.eth.contract(v2_router, abi=uniswapv2_abi)
            V2PoolPath.contract = self.uniswapv2

            v2_factory = w3.to_checksum_address(kwargs.get("v2_factory") or uniswapv2_factory_address)
            self.factoryv2 = self.w3.eth.contract(v2_factory, abi=uniswapv2_factory_abi)

        if self.with_v3:
            self.v3_pool_fees = tuple(kwargs.get("v3_pool_fees") or v3_pool_fees)
            self.v3_pools_fees_x_pivots = tuple(itertools.product(self.pivots, self.v3_pool_fees))

            v3_quoter = w3.to_checksum_address(kwargs.get("v3_quoter") or uniswapv3_quoter_address)
            self.quoter = self.w3.eth.contract(v3_quoter, abi=uniswapv3_quoter_abi)
            V3PoolPath.contract = self.quoter

            v3_factory = w3.to_checksum_address(kwargs.get("v3_factory") or uniswapv3_factory_address)
            self.factoryv3 = self.w3.eth.contract(v3_factory, abi=uniswapv3_factory_abi)

        self.smart_rate_limiter = smart_rate_limiter

    def get_smart_rate_limiter(self) -> Optional[SmartRateLimiter]:
        return self.smart_rate_limiter

    @classmethod
    async def create(
            cls,
            w3: Optional[AsyncWeb3] = None,
            rpc_endpoint: Optional[str] = None,
            with_gas_estimate: bool = False,
            smart_rate_limiter: Optional[SmartRateLimiter] = None) -> "SmartPath":
        """
        Create a SmartPath instance which will search for the best path from v2 and v3 pools.

        :param w3: a valid AsyncWeb3 instance (if no rpc endpoint is given)
        :param rpc_endpoint: an rpc endpoint address (if no w3 instance is given)
        :param with_gas_estimate: Not supported at the moment.
        :param smart_rate_limiter: an instance of SmartRateLimiter to manage rate limits
        :return: a SmartPath instance using v2 and v3 pools
        """
        _w3 = await cls._get_w3(rpc_endpoint, w3)
        chain_id = await _w3.eth.chain_id
        logger.debug(f"Creating SmartPath for V2 and V3 pools on chain id: {chain_id}")
        return cls(_w3, with_gas_estimate, chain_id, True, True, smart_rate_limiter)

    @classmethod
    async def create_v2_only(
            cls,
            w3: Optional[AsyncWeb3] = None,
            rpc_endpoint: Optional[str] = None,
            with_gas_estimate: bool = False,
            smart_rate_limiter: Optional[SmartRateLimiter] = None) -> "SmartPath":
        """
        Create a SmartPath instance which will search for the best path from v2 pools only.

        :param w3: a valid AsyncWeb3 instance (if no rpc endpoint is given)
        :param rpc_endpoint: an rpc endpoint address (if no w3 instance is given)
        :param with_gas_estimate: Not supported at the moment.
        :param smart_rate_limiter: an instance of SmartRateLimiter to manage rate limits
        :return: a SmartPath instance using only v2 pools
        """
        _w3 = await cls._get_w3(rpc_endpoint, w3)
        chain_id = await _w3.eth.chain_id
        logger.debug(f"Creating SmartPath for V2 only pool son chain id: {chain_id}")
        return cls(_w3, with_gas_estimate, chain_id, True, False, smart_rate_limiter)

    @classmethod
    async def create_v3_only(
            cls,
            w3: Optional[AsyncWeb3] = None,
            rpc_endpoint: Optional[str] = None,
            with_gas_estimate: bool = False,
            smart_rate_limiter: Optional[SmartRateLimiter] = None) -> "SmartPath":
        """
        Create a SmartPath instance which will search for the best path from v3 pools only.

        :param w3: a valid AsyncWeb3 instance (if no rpc endpoint is given)
        :param rpc_endpoint: an rpc endpoint address (if no w3 instance is given)
        :param with_gas_estimate: Not supported at the moment.
        :param smart_rate_limiter: an instance of SmartRateLimiter to manage rate limits
        :return: a SmartPath instance using only v3 pools
        """
        _w3 = await cls._get_w3(rpc_endpoint, w3)
        chain_id = await _w3.eth.chain_id
        logger.debug(f"Creating SmartPath for V3 only pools on chain id: {chain_id}")
        return cls(_w3, with_gas_estimate, chain_id, False, True, smart_rate_limiter)

    @classmethod
    async def create_custom(
            cls,
            w3: Optional[AsyncWeb3] = None,
            rpc_endpoint: Optional[str] = None,
            with_gas_estimate: bool = False,
            smart_rate_limiter: Optional[SmartRateLimiter] = None,
            **kwargs: Any) -> "SmartPath":
        """
        Create a SmartPath instance with custom V2 and/or v3 addresses and pivot tokens. Customization is done thanks
        to the following optional keyword arguments:

        * pivot_tokens: Sequence[str] - addresses of the token used for multi-hop pools, like weth, usdc, usdt, dai, ...
        * v3_pool_fees: tuple of v3 fees as basis point. eg: (100, 500, 3000, 10000)
        * v2_router: str - v2 router address
        * v2_factory: str - v2 factory address
        * v3_quoter: str - v3 quoter address
        * v3_factory: str - v3 factory address

        :param w3: a valid AsyncWeb3 instance (if no rpc endpoint is given)
        :param rpc_endpoint: an rpc endpoint address (if no w3 instance is given)
        :param with_gas_estimate: Not supported at the moment.
        :param smart_rate_limiter: an instance of SmartRateLimiter to manage rate limits
        :param kwargs: keyword args to customize V2 and/or V3 pools (see above)
        :return: a custom SmartPath instance
        """
        _w3 = await cls._get_w3(rpc_endpoint, w3)
        _chain_id = await _w3.eth.chain_id
        logger.debug(f"Creating custom SmartPath on chain id: {_chain_id}")

        if kwargs.get("pivot_tokens"):
            pivots = tuple(cast(Sequence[str], kwargs.get("pivot_tokens")))
            _pivots = await cls._get_pivot_tokens(pivots, _w3)
        else:
            _pivots = None

        with_v2 = kwargs.get("v2_router") and kwargs.get("v2_factory")
        with_v3 = kwargs.get("v3_quoter") and kwargs.get("v3_factory")
        if not with_v2 and not with_v3:
            raise SmartPathException("Must provide v2 and/or v3 addresses")

        return cls(
            _w3,
            with_gas_estimate,
            _chain_id,
            with_v2=bool(with_v2),
            with_v3=bool(with_v3),
            smart_rate_limiter=smart_rate_limiter,
            pivot_tokens=_pivots,
            v3_pool_fees=kwargs.get("v3_pool_fees"),
            v2_router=kwargs.get("v2_router"),
            v2_factory=kwargs.get("v2_factory"),
            v3_quoter=kwargs.get("v3_quoter"),
            v3_factory=kwargs.get("v3_factory"),
        )

    @staticmethod
    async def _get_pivot_tokens(pivots: Sequence[str], w3: AsyncWeb3) -> Tuple[Token, ...]:
        pivot_coros = [SmartPath._get_token_at_creation(AsyncWeb3.to_checksum_address(pivot), w3) for pivot in pivots]
        return tuple(await asyncio.gather(*pivot_coros))

    @staticmethod
    async def _get_w3(rpc_endpoint: Optional[str], w3: Optional[AsyncWeb3]) -> AsyncWeb3:
        if w3:
            _w3 = w3
        elif rpc_endpoint:
            _w3 = AsyncWeb3(AsyncHTTPProvider(rpc_endpoint, {"timeout": 5}))
        else:
            raise ValueError("Invalid parameters. Must provide either an AsyncWeb3 instance or an rpc address")
        return _w3

    async def _get_symbol(self, contract: AsyncContract) -> str:
        try:
            return str(await self._contract_function_call(contract.functions.symbol()))
        except (BadFunctionCallOutput, OverflowError):
            return "???"

    @_rate_limit("eth_call")
    async def _contract_function_call(self, contract_function: AsyncContractFunctions) -> Any:
        return await cast(Awaitable[Any], contract_function.call())

    async def _get_token(self, address: ChecksumAddress, w3: AsyncWeb3) -> Token:
        erc20 = w3.eth.contract(address, abi=erc20_abi)
        symbol, decimals = await asyncio.gather(
            self._get_symbol(erc20),
            self._contract_function_call(erc20.functions.decimals())
        )
        return Token(AsyncWeb3.to_checksum_address(address), symbol, decimals)

    @staticmethod
    async def _get_token_at_creation(address: ChecksumAddress, w3: AsyncWeb3) -> Token:
        erc20 = w3.eth.contract(address, abi=erc20_abi)
        symbol, decimals = await asyncio.gather(
            erc20.functions.symbol().call(),
            erc20.functions.decimals().call(),
        )
        return Token(AsyncWeb3.to_checksum_address(address), symbol, decimals)

    async def _v2_pool_exist(self, token0: Token, token1: Token) -> bool:
        try:
            pool_address = await self._contract_function_call(
                self.factoryv2.functions.getPair(
                    token0.address,
                    token1.address
                )
            )
            return AsyncWeb3.is_checksum_address(pool_address) and not is_null_address(pool_address)
        except asyncio.exceptions.TimeoutError:
            return False

    async def _v3_pool_exist(self, token0: Token, token1: Token, fees: int) -> bool:
        try:
            pool_address = await self._contract_function_call(
                self.factoryv3.functions.getPool(
                    token0.address,
                    token1.address,
                    fees
                )
            )
            return AsyncWeb3.is_checksum_address(pool_address) and not is_null_address(pool_address)
        except asyncio.exceptions.TimeoutError:
            return False

    async def _v2_pools_exists_for_pivot_token(self, token0: Token, token1: Token, pivot_token: Token) -> bool:
        pool_1_address, pool_2_address = await asyncio.gather(
            self._contract_function_call(self.factoryv2.functions.getPair(token0.address, pivot_token.address)),
            self._contract_function_call(self.factoryv2.functions.getPair(pivot_token.address, token1.address)),
        )
        return (
                AsyncWeb3.is_checksum_address(pool_1_address)
                and AsyncWeb3.is_checksum_address(pool_2_address)
                and not is_null_address(pool_1_address)
                and not is_null_address(pool_2_address)
        )

    async def _build_v2_path_list(self, token_in: Token, token_out: Token) -> List[V2PoolPath]:
        v2_path_list: List[V2PoolPath] = []
        if not self.with_v2:
            return v2_path_list

        v2_pools_exist_cor_list = [self._v2_pool_exist(token_in, token_out)]
        filtered_pivots = [pivot for pivot in self.pivots if pivot not in (token_in, token_out)]
        for pivot_token in filtered_pivots:
            v2_pools_exist_cor_list.append(self._v2_pools_exists_for_pivot_token(token_in, token_out, pivot_token))

        v2_pools_exist = await asyncio.gather(*v2_pools_exist_cor_list)

        if v2_pools_exist[0]:
            v2_path_list.append(V2PoolPath((V2OrderedPool(token_in, token_out),), self.smart_rate_limiter))
        for i, result in enumerate(v2_pools_exist[1:]):
            if result:
                v2_path_list.append(
                    V2PoolPath(
                        (V2OrderedPool(token_in, filtered_pivots[i]), V2OrderedPool(filtered_pivots[i], token_out)),
                        self.smart_rate_limiter,
                    )
                )

        return v2_path_list

    async def _get_v3_base_pools(self, token: Token, is_token_in: bool) -> List[V3OrderedPool]:
        v3_pools_exist_cor_list = []
        filtered__v3_pools_fees_x_pivots = [pf for pf in self.v3_pools_fees_x_pivots if pf[0] != token]
        for pivot, fees in filtered__v3_pools_fees_x_pivots:
            if pivot != token:
                v3_pools_exist_cor_list.append(self._v3_pool_exist(token, pivot, fees))

        v3_pools_exist = await asyncio.gather(*v3_pools_exist_cor_list)
        v3_pool_list = []
        for i, result in enumerate(v3_pools_exist):
            if result:
                pivot = filtered__v3_pools_fees_x_pivots[i][0]
                fees = filtered__v3_pools_fees_x_pivots[i][1]
                pool = V3OrderedPool(token, fees, pivot) if is_token_in else V3OrderedPool(pivot, fees, token)
                v3_pool_list.append(pool)

        return v3_pool_list

    async def _get_v3_one_hop_pools(self, token_in: Token, token_out: Token) -> List[V3OrderedPool]:
        v3_pools_exist_cor_list = []
        for fees in self.v3_pool_fees:
            v3_pools_exist_cor_list.append(self._v3_pool_exist(token_in, token_out, fees))
        v3_pools_exist = await asyncio.gather(*v3_pools_exist_cor_list)
        v3_pool_list = []
        for i, result in enumerate(v3_pools_exist):
            if result:
                v3_pool_list.append(V3OrderedPool(token_in, self.v3_pool_fees[i], token_out))
        return v3_pool_list

    async def _build_v3_path_list(self, token_in: Token, token_out: Token) -> List[V3PoolPath]:
        v3_path_list: List[V3PoolPath] = []
        if not self.with_v3:
            return v3_path_list

        one_hop_pools, token_in_base_pools, token_out_base_pools = await asyncio.gather(
            self._get_v3_one_hop_pools(token_in, token_out),
            self._get_v3_base_pools(token_in, True),
            self._get_v3_base_pools(token_out, False),
        )

        for pool in one_hop_pools:
            v3_path_list.append(V3PoolPath((pool,), self.smart_rate_limiter))

        if len(token_in_base_pools) > 0 and len(token_out_base_pools) > 0:
            product = itertools.product(token_in_base_pools, token_out_base_pools)
            two_hop_pools = [p for p in product if p[0].token_out == p[1].token_in]
            for two_hop_pool in two_hop_pools:
                v3_path_list.append(V3PoolPath(two_hop_pool, self.smart_rate_limiter))

        return v3_path_list

    @staticmethod
    def _filter_irrelevant_low_values(
            mixed_weighted_paths: List[MixedWeightedPath],
            best_value: Wei) -> List[MixedWeightedPath]:
        return [
            path
            for path in mixed_weighted_paths
            if path.total_value > best_value * irrelevant_value_filter_multiplier
        ]

    @staticmethod
    def _get_all_mixed_path(
            lower_value_path: MixedWeightedPath,
            higher_value_path: MixedWeightedPath) -> List[MixedWeightedPath]:
        all_paths = []
        for low_weight, high_weight in weight_combinations:
            lower_weighted_path = WeightedPath(
                router_function=lower_value_path.weighted_paths[0].router_function,
                pool_path=lower_value_path.weighted_paths[0].pool_path,
                weight=low_weight,
            )
            higher_weighted_path = WeightedPath(
                router_function=higher_value_path.weighted_paths[0].router_function,
                pool_path=higher_value_path.weighted_paths[0].pool_path,
                weight=high_weight,
            )
            all_paths.append(MixedWeightedPath((lower_weighted_path, higher_weighted_path)))
        return all_paths

    async def get_swap_in_path(
            self,
            amount: Wei,
            token_in_address: ChecksumAddress,
            token_out_address: ChecksumAddress) -> Tuple[WeightedPathResult, ...]:
        token_in, token_out = await asyncio.gather(
            self._get_token(token_in_address, self.w3),
            self._get_token(token_out_address, self.w3),
        )
        v2_pool_paths, v3_pool_paths = await asyncio.gather(
            self._build_v2_path_list(token_in, token_out),
            self._build_v3_path_list(token_in, token_out),
        )

        v2_mixed_paths = [
            MixedWeightedPath(
                (WeightedPath(RouterFunction.V2_SWAP_EXACT_IN, pool_path, 100), )
            ) for pool_path in v2_pool_paths
        ]
        v3_mixed_paths = [
            MixedWeightedPath(
                (WeightedPath(RouterFunction.V3_SWAP_EXACT_IN, pool_path, 100), )
            ) for pool_path in v3_pool_paths
        ]

        computing_value_coros = [path.compute_path_values(amount) for path in v2_mixed_paths]
        computing_value_coros.extend([path.compute_path_values(amount) for path in v3_mixed_paths])
        await asyncio.gather(*computing_value_coros)

        v2_mixed_paths.sort(key=lambda mp: mp.total_value, reverse=True)
        v3_mixed_paths.sort(key=lambda mp: mp.total_value, reverse=True)

        best_value = max(
            v2_mixed_paths[0].total_value if len(v2_mixed_paths) > 0 else 0,
            v3_mixed_paths[0].total_value if len(v3_mixed_paths) > 0 else 0,
        )

        v2_mixed_paths = self._filter_irrelevant_low_values(v2_mixed_paths, Wei(best_value))
        logger.debug(f"V2 Paths: {v2_mixed_paths}")
        v3_mixed_paths = self._filter_irrelevant_low_values(v3_mixed_paths, Wei(best_value))
        logger.debug(f"V3 Paths: {v3_mixed_paths}")

        if len(v2_mixed_paths) == len(v3_mixed_paths) == 0:
            return ()
        elif len(v3_mixed_paths) == 0:
            return v2_mixed_paths[0].output()
        elif len(v2_mixed_paths) == 0:
            return v3_mixed_paths[0].output()
        else:
            if v2_mixed_paths[0].total_value > v3_mixed_paths[0].total_value:
                lower_value_path = v3_mixed_paths[0]
                higher_value_path = v2_mixed_paths[0]
            else:
                lower_value_path = v2_mixed_paths[0]
                higher_value_path = v3_mixed_paths[0]

            all_mixed_paths = self._get_all_mixed_path(lower_value_path, higher_value_path)
            computing_value_coros = [path.compute_path_values(amount) for path in all_mixed_paths]
            await asyncio.gather(*computing_value_coros)
            all_mixed_paths.extend([v2_mixed_paths[0], v3_mixed_paths[0]])
            all_mixed_paths.sort(key=lambda mp: mp.total_value, reverse=True)
            logger.debug(f"All mixed paths: {all_mixed_paths}")

            return all_mixed_paths[0].output()



from typing import Any

from web3 import Web3
from web3.types import Wei


def to_wei(amount: Any) -> Wei:
    return Wei(int(amount))


def is_null_address(address: Any) -> bool:
    try:
        return True if Web3.to_checksum_address(address) and "0" * 40 in address else False
    except (TypeError, ValueError):
        return False



from functools import wraps
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    Protocol,
    TypedDict,
)

from credit_rate_limit import (
    CountRateLimiter,
    CreditRateLimiter,
    throughput,
)
from credit_rate_limit.rate_limiter import DecoratedSignature


class MethodCredit(TypedDict):
    eth_call: int


class SmartRateLimiter:
    def __init__(
            self,
            interval: float,
            max_count: Optional[int] = None,
            max_credits: Optional[int] = None,
            method_credits: Optional[MethodCredit] = None,
            ) -> None:
        self.interval = interval
        self.max_count = max_count
        self.max_credits = max_credits
        self.method_credits = method_credits
        if self.max_credits:
            if self.method_credits:
                self.rate_limiter = CreditRateLimiter(max_credits, interval)
            else:
                raise ValueError("Missing parameter 'method_credits' needed for credit rate limit")
        elif self.max_count:
            self.rate_limiter = CountRateLimiter(max_count, interval)
        else:
            raise ValueError("Missing parameter: either 'max_count' or 'max_credits' (and 'method_credits') is needed")


class GotSmartRateLimiter(Protocol):
    def get_smart_rate_limiter(self) -> Optional[SmartRateLimiter]: ...


def _rate_limit(method_name: Literal["eth_call"]) -> Callable[[Callable[..., Any]], Any]:
    def decorator(func: DecoratedSignature) -> Any:
        @wraps(func)
        def wrapper(self_: GotSmartRateLimiter, *args: Any, **kwargs: Any) -> Any:
            smart_rate_limiter = self_.get_smart_rate_limiter()
            if smart_rate_limiter:
                rate_limiter = smart_rate_limiter.rate_limiter
                if isinstance(rate_limiter, CreditRateLimiter) and smart_rate_limiter.method_credits:
                    request_credits = smart_rate_limiter.method_credits[method_name]
                    return throughput(rate_limiter, request_credits=request_credits)(func)(self_, *args, **kwargs)
                elif isinstance(rate_limiter, CountRateLimiter):
                    return throughput(rate_limiter)(func)(self_, *args, **kwargs)
                else:
                    raise ValueError("Missing parameter 'method_credits' , or unknown rate limiter")
            else:
                return func(self_, *args, **kwargs)
        return wrapper
    return decorator



from typing import (
    Dict,
    Tuple,
)

from web3 import Web3

from ._datastructures import Token


erc20_abi = '[{"constant":true,"inputs":[],"name":"name","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_spender","type":"address"},{"name":"_value","type":"uint256"}],"name":"approve","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"totalSupply","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transferFrom","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"decimals","outputs":[{"name":"","type":"uint8"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"}],"name":"balanceOf","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"symbol","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transfer","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"},{"name":"_spender","type":"address"}],"name":"allowance","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},{"indexed":true,"name":"_to","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Transfer","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_spender","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Approval","type":"event"}]'  # noqa
uniswapv2_abi = '[{"inputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"},{"internalType":"address[]","name":"path","type":"address[]"}],"name":"getAmountsOut","outputs":[{"internalType":"uint256[]","name":"amounts","type":"uint256[]"}],"stateMutability":"view","type":"function"}]'  # noqa
uniswapv2_address = Web3.to_checksum_address("0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D")
uniswapv2_factory_abi = '[{"constant":true,"inputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"address","name":"","type":"address"}],"name":"getPair","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"}]'  # noqa
uniswapv2_factory_address = Web3.to_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f")

uniswapv3_quoter_address = Web3.to_checksum_address("0x61fFE014bA17989E743c5F6cB21bF9697530B21e")
uniswapv3_quoter_abi = '[{"inputs":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"uint256","name":"amountIn","type":"uint256"}],"name":"quoteExactInput","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"},{"internalType":"uint160[]","name":"sqrtPriceX96AfterList","type":"uint160[]"},{"internalType":"uint32[]","name":"initializedTicksCrossedList","type":"uint32[]"},{"internalType":"uint256","name":"gasEstimate","type":"uint256"}],"stateMutability":"nonpayable","type":"function"}]'  # noqa
uniswapv3_factory_address = Web3.to_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984")
uniswapv3_factory_abi = '[{"inputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"address","name":"","type":"address"},{"internalType":"uint24","name":"","type":"uint24"}],"name":"getPool","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"}]'  # noqa

pivot_tokens: Dict[int, Tuple[Token, ...]] = {
    1: (  # Ethereum
        Token(Web3.to_checksum_address("0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48"), "USDC", 6),
        Token(Web3.to_checksum_address("0xdac17f958d2ee523a2206206994597c13d831ec7"), "USDT", 6),
        Token(Web3.to_checksum_address("0x6b175474e89094c44da98b954eedeac495271d0f"), "DAI", 18),
        Token(Web3.to_checksum_address("0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2"), "WETH", 18),
    ),
    1337: (  # Ganache for Ethereum
        Token(Web3.to_checksum_address("0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48"), "USDC", 6),
        Token(Web3.to_checksum_address("0xdac17f958d2ee523a2206206994597c13d831ec7"), "USDT", 6),
        Token(Web3.to_checksum_address("0x6b175474e89094c44da98b954eedeac495271d0f"), "DAI", 18),
        Token(Web3.to_checksum_address("0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2"), "WETH", 18),
    ),
}

v3_pool_fees = (100, 500, 3000, 10000)
weight_combinations = tuple((i, j) for i in range(10, 100, 10) for j in range(10, 100, 10) if i + j == 100 and i <= j)

irrelevant_value_filter_multiplier = 0.9



import pytest
from web3.types import Wei

from uniswap_smart_path._utilities import (  # noqa
    is_null_address,
    to_wei,
)

from .conftest import tokens


@pytest.mark.parametrize(
    "amount, expected_result",
    (
        (10**18, Wei(10**18)),
        (123.45, Wei(123)),
    )
)
def test_to_wei(amount, expected_result):
    assert expected_result == to_wei(amount)


@pytest.mark.parametrize(
    "address, expected_result",
    (
        (tokens["UNI"], False,),
        ("0"*40, True),
        ("Not an address", False),
        (None, False),
    )
)
def test_is_null_address(address, expected_result):
    assert expected_result == is_null_address(address)
